
/**
 * pcx.c: Functions for reading pcx files.
 * ====================================================================
 * Copyright (c) 2007 Jonathan Coome.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * ====================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>

#include "pcx.h"
#include "iobuffer.h"


static pcx_header *read_header(FILE *file)
{
	struct pcx_header *header;
	int size = sizeof(struct pcx_header);

	header = malloc(size);
	if (header == NULL) return NULL;

	fread(header, size, 1, file);
	return header;
}

static uint8_t *read_palette(FILE *file, pcx_header *header)
{
	uint8_t palette[769];
	uint8_t *palette_p;

	if (header->bits_per_pixel != 8 || header->colour_planes != 1) {
		/* This image doesn't have a palette at the end of the file. */
		return NULL;
	}

	fseek(file, -769, SEEK_END);

	fread(palette, sizeof(palette), 1, file);
	if (palette[0] != 12) return NULL;
	
	palette_p = malloc(768);
	if (palette_p == NULL) return NULL;

	memcpy(palette_p, &palette[1], 768);
	return palette_p;
}

static void *read_image_cleanup(pcx_image *image, uint8_t *line, iobuffer *buf)
{
	free(buf);
	free(line);
	pcx_free_image(image);

	return NULL;
}

pcx_image *pcx_read_image(FILE *file)
{
	int i;
	int total_bytes;
	int image_size;

	uint8_t *line;
	iobuffer *buf;
	pcx_image *image;
	pcx_header *header;
	
	image = calloc(sizeof(pcx_image), 1);
	if (image == NULL) return NULL;

	header = read_header(file);
	if (header == NULL) return read_image_cleanup(image, NULL, NULL);

	image->header = header;
	image->x_size = header->x_max - header->x_min + 1;
	image->y_size = header->y_max - header->y_min + 1;
 	image_size = image->x_size * image->y_size * header->colour_planes;

	image->image = malloc(image_size);
	if (image->image == NULL) return read_image_cleanup(image, NULL, NULL);

 	total_bytes = header->colour_planes * header->bytes_per_line;

	line = malloc(total_bytes);
	if (line == NULL) return read_image_cleanup(image, NULL, NULL);

	buf = alloc_buffer(file);
	if (buf == NULL) return read_image_cleanup(image, line, NULL);

	for (i = 0; i < image->y_size; i++) {
		int written = 0;
		uint8_t *line_p = line;
		int length = 0;

		while (written < total_bytes) {
			uint8_t byte = read_byte(buf);
			
			uint8_t colour;

			if ((byte & 0xC0) == 0xC0) {
				length = (byte & 0x3F);
				colour = read_byte(buf);
			} else {
				length = 1;
				colour = byte;
			}

			for (; length > 0 && written < total_bytes; length--) {
				*line_p = colour;
				line_p++;
				written++;
			}
		}

		memcpy(image->image + i * image->x_size, line, image->x_size);
	}

	free(line);
	free_buffer(buf);

	image->palette = read_palette(file, header);

	return image;
}

void pcx_free_image(pcx_image *image)
{
	free(image->header);
	free(image->image);
	free(image->palette);
	free(image);
}
